from subprocess import Popen, PIPE
import os
import urllib
import tempfile
import zipfile
import shutil
import stat
import logging
import logging.handlers
import logger

from config import SamConfig
from models import AppVersions
from storage import Applications, Installations
import runner


default_config_path = '/usr/local/sam/config'


def get_sam(config_path=default_config_path):

    config = SamConfig(os.path.join(config_path, 'sam.cfg'))

    apps_dir = config.apps_dir()
    status_dir = config.status_dir()

    app_index_filename = os.path.join(apps_dir, 'index')
    app_status_filename = os.path.join(status_dir, 'installed.apps')

    installations = Installations(app_status_filename)
    applcations = Applications(app_index_filename)

    manager = Manager(config, applcations, installations)

    return manager


class Manager:

    def __init__(self, config, applications, installations):
        self.config = config
        self.apps_dir = config.apps_dir()

        self.applications = applications
        self.installations = installations
        self.logger = logger.get_logger('manager')

    def update(self, release=None):
        self.logger.info("update")
        release_filename = os.path.join(self.config.status_dir(), 'release')
        if not release:
            if os.path.isfile(release_filename):
                with open(release_filename, 'r') as f:
                    release = f.read()
            else:
                raise Exception('The release is not set')

        if release:
            apps_url_template = self.config.apps_url_template()
            apps_url = apps_url_template.format(release)

            extract_dir = tempfile.mkdtemp()
            archive_filename = os.path.join(extract_dir, 'archive.zip')

            urllib.urlretrieve(apps_url, filename=archive_filename)

            extract_to_dir = os.path.join(extract_dir, 'unzipped')

            z = zipfile.ZipFile(archive_filename, 'r')
            z.extractall(extract_to_dir)

            subdirs = [name for name in os.listdir(extract_to_dir) if os.path.isdir(os.path.join(extract_to_dir, name))]

            extracted_apps_dir = os.path.join(extract_to_dir, subdirs[0])

            for filename in [name for name in os.listdir(extracted_apps_dir) if os.path.isfile(os.path.join(extracted_apps_dir, name))]:
                full_filename = os.path.join(extracted_apps_dir, filename)
                st = os.stat(full_filename)
                os.chmod(full_filename, st.st_mode | stat.S_IEXEC)

            shutil.rmtree(self.apps_dir, ignore_errors=True)
            shutil.copytree(extracted_apps_dir, self.apps_dir)

            with open(release_filename, 'w+') as f:
                f.write(release)

            return self.__upgradable_apps()

    def newer_available(self, application):
        if application.installed_version:
            return application.current_version != application.installed_version
        else:
            return application.app.required

    def __upgradable_apps(self):
        return [a for a in self.list() if self.newer_available(a)]

    def upgrade_all(self):
        applications = self.__upgradable_apps()
        for application in applications:
            if application.installed_version:
                self.upgrade(application.app.id)
            else:
                self.install(application.app.id)

    def list(self):
        apps = self.applications.list()
        return [self.get_app_versions(app) for app in apps]

    def remove(self, app_id):
        a = self.get_app(app_id, True)
        self.run_hook(a.app, 'pre-remove')
        self.run(a.app, 'remove')
        self.installations.remove(a.app)
        return "removed successfully"

    def install(self, app_id):
        a = self.get_app(app_id, False)
        app = a.app
        self.run(app, 'install')
        self.installations.update(app, a.current_version)
        return "installed successfully"

    def upgrade(self, app_id):
        a = self.get_app(app_id, True)
        app = a.app
        self.run(app, 'upgrade')
        self.installations.remove(app)
        self.installations.update(app, a.current_version)
        return "upgraded successfully"

    def verify(self, app_id):
        a = self.get_app(app_id, False)
        self.run(a.app, 'verify')
        return "verified successfully"

    def get_app_versions(self, app):
        version = self.app_version(app)
        installed_version = self.installations.version(app)
        return AppVersions(app, version, installed_version)

    def app_version(self, app):
        version_cmd = "{} version".format(self.get_script(app))
        version = Popen(version_cmd, shell=True, stdout=PIPE).stdout.read().strip()
        return version

    def get_app(self, app_id, installed=False):
        found = next((a for a in self.list() if a.app.id == app_id), None)

        if not found:
            raise Exception("no such app: {}".format(app_id))

        if not (found.installed_version is not None) == installed:
            if installed:
                raise Exception('{} should be installed but it is not'.format(app_id))
            else:
                raise Exception('{} should not be installed but it is'.format(app_id))

        return found

    def run(self, app, action):
        if not runner.call(self.app_command(app, action), self.logger, stdout_log_level=logging.INFO, shell=True) == 0:
            raise Exception("unable to {} {}".format(action, app.id))

    def app_command(self, app, command):
        full_filename = self.get_script(app)
        cmd = ' '.join([full_filename, command])
        return cmd

    def get_script(self, app):
        return os.path.join(self.apps_dir, app.id)

    def run_hook(self, app, action):

        hook_bin = os.path.join(self.config.bin_dir(), app.id + '-' + action)
        if not os.path.isfile(hook_bin):
            self.logger.info("{} hook is not found, skipping".format(hook_bin))
            return

        if not runner.call(hook_bin, self.logger, stdout_log_level=logging.INFO, shell=True) == 0:
            raise Exception("unable to run {}".format(hook_bin))
